use std::collections::{HashMap, HashSet};
use std::iter::FromIterator;

const FILE_STR: &'static str = include_str!("input.txt");

#[derive(Debug, Default)]
struct ScratchCard {
    winning: Vec<usize>,
    chosen: Vec<usize>,
    chosen_winners: HashSet<usize>,
}

impl ScratchCard {
    fn new(winning: Vec<usize>, chosen: Vec<usize>) -> Self {
        ScratchCard {
            winning,
            chosen,
            ..Self::default()
        }
    }

    fn find_chosen_winners(&mut self) {
        self.winning
            .iter()
            .filter(|w| self.chosen.contains(w))
            .for_each(|cw| _ = self.chosen_winners.insert(*cw));
    }

    fn score(&mut self) -> usize {
        self.find_chosen_winners();
        if self.chosen_winners.is_empty() {
            return 0;
        }

        let winner_count = self.chosen_winners.len() as u32;
        usize::pow(2, winner_count - 1)
    }
}

#[derive(Debug, Default)]
struct GameMap(HashMap<usize, ScratchCard>);

impl GameMap {
    fn parse(table: &str) -> Self {
        Self(HashMap::from_iter(
            table
                .split('\n')
                .filter(|s| s.len() > 0)
                .map(parse_game_table_line),
        ))
    }

    fn find_all_chosen_winners(&mut self) {
        self.0.iter_mut().for_each(|(_, sc)| {
            sc.find_chosen_winners();
        })
    }

    fn get_score_sum(&mut self) -> usize {
        self.0.iter_mut().map(|(_, sc)| sc.score()).sum()
    }
}

fn parse_game_table_line(line: &str) -> (usize, ScratchCard) {
    let parts: Vec<&str> = line.split(": ").collect();
    let card_num_parts = parts[0].replace("Card", "");
    let card_num = parse_num_str(&card_num_parts)[0];
    let nums: Vec<Vec<usize>> = parts[1].split(" | ").map(parse_num_str).collect();

    (card_num, ScratchCard::new(nums[0].clone(), nums[1].clone()))
}

/// Parse numbers from a line, separated/surrounded by whitespace
fn parse_num_str(s: &str) -> Vec<usize> {
    s.split_ascii_whitespace()
        .filter(|s| s.len() > 0)
        .map(|s| s.trim())
        .map(|s| usize::from_str_radix(s, 10).unwrap())
        .collect::<Vec<usize>>()
}

fn part_one() {
    let mut map = GameMap::parse(FILE_STR);
    let sum = map.get_score_sum();
    println!("Part 1 Score Sum {}", sum);
}

fn main() {
    part_one();
}

#[cfg(test)]
mod tests {
    use super::*;
    const EXAMPLE_FILE_STR: &'static str = include_str!("example_input.txt");

    #[test]
    fn test_part1_example() {
        let mut map = GameMap::parse(EXAMPLE_FILE_STR);
        let actual = map.get_score_sum();
        assert_eq!(13, actual);
    }
}
